#!/usr/bin/env python
# coding: utf-8

# In[1]:


import at
import at.plot
import at.tracking
import numpy as np
import matplotlib.pyplot as plt

ring = at.load_mat('./dba.mat', mat_key='RING')
arc = at.load_mat('./arc.mat', mat_key='ARC')


# ## Simple Tracking, Optics and Orbit

# Now we will do some simple tracking through the lattice, and also show how to compute the optics and the orbit. 
# 
# First we need some particles to track, these must be in the format of ((6,N)) where 6 is the number of dimensions of AT (x,x',y,y',delta,ct) and N is the number of particles. This array can be manually created, but it is also desirable to generate a beam that is matched to the lattice. We can do this by generating the sigma correlation matrix using 'sigma_matrix'. The function will return a correlation matrix that can be used to generate a matched beam for the lattice you provide it. However there are some interesting things to be aware of, e.g.

# In[2]:


ring.disable_6d()
#sig_mat = at.sigma_matrix(ring)


# We have give a lattice with radiation off and asked for a correlation matrix. However, the emittances of a lattice with no radiation are 0 in x and y. Therefore we must provide the emittances that we want.

# In[3]:


sig_mat = at.sigma_matrix(ring, emitx=100e-12, emity=1e-12)
sig_mat


# Due to the fact that the rad is off. We only have correlations for x,x' and y,y' and nothing in the longitudinal plane. But we can switch rad on and compare the difference. Here you may notice that we don't need to give the emittances, that is because if radiation is on and no emittances are specified, it will take the result from envelope_parameters. If you want a beam with veritcal emittance, you can also still specify it. There are many different cases that can be seen in the help, but the exception that is raised if something is missing is quite explanatory. 

# In[4]:


sig_mat_rad_on = at.sigma_matrix(ring.radiation_on(copy=True), emity=10e-12)
sig_mat_rad_on


# Now that we have the matrix we can generate a beam using 'beam'.

# In[5]:


nparts = 50000
parts = at.beam(nparts, sig_mat_rad_on)
print(parts.shape)


# In[6]:


coord=5
print(np.std(parts[coord,:]))
plt.hist(parts[coord,:], bins=100)


# Now let's do some tracking. but to make it easier we can start with only a single particle.

# In[7]:


p_in = np.array([1e-3, 0., 0., 0., 0., 0.]) #must be floats not ints
p_in.shape


# We need to use the function lattice_pass. We must define the refpts in the lattice where we want to save the output. And we also want to specify nturns. There are two ways of tracking, either you use nturns=1 within a for loop, this has the benefit of allowing you to perform your own TbT analysis and save relevant parameters only. Or you can tracking for many turns, look at the full coordinates of the whole tracking, and perform the post processing later.

# In[8]:


#get_ipython().run_line_magic('pinfo', 'ring.track')


# In[9]:


# Method 1, the output of the tracking is a big array with shape (6, N, R, T)
# where 6 is the particle coordinate, N is the particle index, R is the element index, 
# and T is the turn number
pout, *_ = at.lattice_pass(ring,p_in, refpts=np.arange(len(ring)), nturns=200)


# In[10]:


tunex, tuney = ring.get_tune()
xdat = pout[0,0,:]
fig, (ax1,ax2) = plt.subplots(2,1)
ax1.plot(xdat)
ax2.plot(np.fft.rfftfreq(len(xdat)), np.abs(np.fft.rfft(xdat)))
ax2.axvline(tunex, color='r', linestyle='dashed')
plt.show()


# An important note, back to pointers again! If you prefer to use them in a turn by turn way, you can go like this

# In[11]:


p_in = np.zeros((6,3))
p_in[0,:] = np.array([1e-4,1e-3,3e-3])
n_turns=200
x_data = np.zeros((p_in.shape[1], n_turns))
xp_data = np.zeros((p_in.shape[1], n_turns))
for i in np.arange(n_turns):
    _ = at.lattice_pass(ring,p_in, nturns=1, refpts=None) #in_place means modify input particle, refpts=None stops the track function from generating big arrays for unneeded output
    x_data[:, i] = p_in[0,:]
    xp_data[:, i] = p_in[1,:]


# In[12]:


plt.plot(x_data[0,:], xp_data[0,:], linestyle='None', marker='.', label='1e-4')
plt.plot(x_data[1,:], xp_data[1,:], linestyle='None', marker='.', label='1e-3')
plt.plot(x_data[2,:], xp_data[2,:], linestyle='None', marker='.', label='3e-3')
plt.legend()
plt.xlabel('x [m]')
plt.ylabel('xp')


# The following section is very important. As we saw, we have the ability to enable or disable 6d motion. When computing the linear optics or the orbit, the 'get_optics' and 'find_orbit' functions will automatically select the most appropriate method to use out of the methods available. 
# 
# > at.linopt2
# 
# > at.linopt4 / at.find_orbit4
# 
# > at.linopt6 / at.find_orbit6
# 
# for 2, 4 and 6 dimensional tracking respectively. One can also specifically call other methods in order to speed up computation (for example if you are doing uncoupled matching, it is faster to call at.linopt2 rather than do the full 6d computation)

# In[13]:


ring.disable_6d()
s_positions = ring.get_s_pos(refpts=np.arange(len(ring)))
orbit0_4, orbit_4 = ring.find_orbit(refpts=np.arange(len(ring)))
ring.enable_6d()
orbit0_6, orbit_6 = ring.find_orbit(refpts=np.arange(len(ring)))
print(orbit_4.shape)


# In[14]:


plt.plot(s_positions, orbit_4[:,0], label='find_orbit4')
plt.plot(s_positions, orbit_6[:,0], label='find_orbit6')
plt.legend()
plt.xlabel('s [m]')
plt.ylabel('x [m]')
plt.show()


# In[15]:


#get_ipython().run_line_magic('pinfo', 'at.get_optics')


# In[16]:


ring.disable_6d()
ld0_4d, bd_4d, ld_4d = at.get_optics(ring, refpts=np.arange(len(ring)), method=at.linopt2)
ring.enable_6d()
ld0_6d, bd_6d, ld_6d = at.get_optics(ring, refpts=np.arange(len(ring)))
ld0_6db, bd_6db, ld_6db = at.get_optics(ring, refpts=np.arange(len(ring)))


# In[17]:


plt.plot(ld_4d.beta[:,1])
plt.plot(ld_6d.beta[:,1])
plt.show()
fig, ax = plt.subplots()
axb = ax.twinx()
ax.plot(s_positions, ld_6d.beta[:,1] - ld_4d.beta[:,1], color='r')
axb.plot(s_positions, orbit_6[:,0], label='find_orbit6', color='b')

plt.show()


# In[18]:


#get_ipython().run_line_magic('pinfo', 'at.find_orbit4')


# # Orbit Correction

# By now we are basically experts. What we will now do is show how to use what we have learned and do some orbit correction.
# First we need to get the orbit response matrix. This is a big matrix showing the orbit response from each steerer action.

# In[19]:


ring.disable_6d()
corrector_refpts = at.get_uint32_index(ring, 'Corr*')
bpm_refpts = at.get_uint32_index(ring, 'BPM*')

orm = np.zeros((len(bpm_refpts), len(corrector_refpts)))


# In[20]:


kickang = 1e-6
for ic, corrector_ind in enumerate(corrector_refpts):
    #init_steer_value = ring[corrector_ind].KickAngle[1]
    #_, ref_orbit = ring.find_orbit(refpts=bpm_refpts)
    ring[corrector_ind].KickAngle[1] += kickang
    _, pos_orbit = ring.find_orbit(refpts=bpm_refpts)
    ring[corrector_ind].KickAngle[1] -= 2*kickang
    _, neg_orbit = ring.find_orbit(refpts=bpm_refpts)
    ring[corrector_ind].KickAngle[1] += kickang
    orm[:, ic] = (pos_orbit[:,2] - neg_orbit[:,2])/kickang


# In[21]:


plt.pcolormesh(orm)
plt.xlabel('corrector #')
plt.ylabel('bpm #')


# Now we can put a bad orbit, we can do this by introducing a misalignment on a couple of quadrupoles. Possible functions for this are 'set_shift' and 'set_tilt'

# In[22]:


ring_err = ring.deepcopy()
'''
at.set_shift?
quad_msk = at.get_bool_index(ring_err, at.Quadrupole)
shift_vals = np.zeros(quad_msk.shape[0]-1)
shift_vals[np.where(quad_msk)[0][0]] = 1e-6
shift_vals[np.where(quad_msk)[0][3]] = 5e-6
shift_vals[np.where(quad_msk)[0][4]] = -2e-6
shift_vals[np.where(quad_msk)[0][9]] = 3e-6
shift_vals[np.where(quad_msk)[0][11]] = 10e-6
shift_vals[np.where(quad_msk)[0][17]] = 0.5e-6
at.set_shift(ring_err, np.zeros(len(shift_vals)), shift_vals)
'''
all_corr_vals = at.get_value_refpts(ring_err, 'Corr*', 'KickAngle', 1)
all_corr_vals[0] = 2e-6
all_corr_vals[4] = -3.5e-5
all_corr_vals[7] = 6e-6

at.set_value_refpts(ring_err, 'Corr*', 'KickAngle', all_corr_vals, 1)


# In[23]:


orb0, orb_bad = ring_err.find_orbit(refpts=bpm_refpts)
plt.plot(orb_bad[:,2])
plt.show()


# In[24]:


U,W,V = np.linalg.svd(orm, full_matrices=False)
Utran = np.transpose(U)
Winv = np.linalg.inv(np.diag(W))
Vtran = np.transpose(V)


# In[25]:


niters = 10
orb_to_correct = orb_bad
for it in np.arange(niters):
    dk = np.dot(Vtran, np.dot(Winv, np.dot(Utran, orb_to_correct[:,2])))
    print(np.sum(dk))
    at.set_value_refpts(ring_err, 'Corr*', 'KickAngle', -dk, 1)
    orb0, orb_to_correct = ring_err.find_orbit(refpts=bpm_refpts)


# In[26]:


plt.plot(orb_bad[:,2], 'r', label='before')
plt.plot(orb_to_correct[:,2], 'b', label='after')
plt.legend()
plt.show()

plt.plot(all_corr_vals, 'r', label='before')
plt.plot(at.get_value_refpts(ring_err, 'Corr*', 'KickAngle', 1), 'b', label='after')
plt.legend()
plt.show()


# ### MATCHING

# So we want to do some matching! The final piece of the introductory puzzle. Lets some back to our arc cell. Then lets ruin it a bit for fun in order to give us something to match back to.

# In[27]:


arc.plot_beta()


# We can put some errors to make it a bit messy, then rematch to these values. However, we can use transfer line mode to speed up the matching and use only half of the cell as it is symmetric. First we need to save the initial optics, which we can use as input into the matching

# In[28]:


ld0, bd, ld = at.get_optics(arc)


# In[29]:


ld0


# In[30]:


half_arc = arc[:at.get_uint32_index(arc, 'BPM_CellCenter')[0]]
print(ld0.dispersion)
#ld0.dispersion[0] = 1e-1
print(ld0.dispersion)

#This uses plot_beta in transfer_line mode. When you provide twiss_in, you set
# the initial optics functions and then progress them in one pass through your lattice
arc.plot_beta(twiss_in=ld0)


# In[31]:


half_arc


# In[32]:


import random
quad_refs = at.get_uint32_index(half_arc, at.Quadrupole)
sext_refs = at.get_uint32_index(half_arc, at.Sextupole)
for quad_ind in quad_refs:    
    half_arc[quad_ind].K *= 1 - random.randint(-50,50)/100
    
for sext_ind in sext_refs:    
    half_arc[sext_ind].H *= 1 - random.randint(-50,50)/100


# In[33]:


half_arc.plot_beta(twiss_in=ld0)


# Now we have a messed up lattice, some initial optics parameters, and a rough idea on what we want it to look like again. So lets try and match it.
# 
# We must define our Constraints (or Goals) and our Variables.
# 
# Possible Constraint containers:
# 
# Constraints - Generic container for user defined functions<br>
# ElementConstraints - Constraints related to attributes of lattice elements <br>
# EnvelopeConstraints - Constraints for values that can be found with envelope_parameters <br>
# LinoptConstraints - Constraints for values that can be found with linopt output <br>
# OrbitConstraints - Constraints relating to the orbit
# 
# Possible Variable containers:
# 
# Variable - Generic container for user defined variables <br>
# ElementVariable - Variable related to attributes of lattice elements <br> 
# 
# We can start with some simple linear optics constraints, which we can recover using the quadrupoles and sextupoles strengths

# In[34]:


#get_ipython().run_line_magic('pinfo', 'at.Variable')


# In[35]:


# First we initialise the constraint type by passing the ring and information about the setup
# These are the arguments that are passed to linopt when computing the optics
constr = at.LinoptConstraints(half_arc, dp=0.0, coupled=False, twiss_in=ld0)


# In[36]:


#Now we can add our optics constraints.
# Weights can also be applied here. See the help of constr.add. Also if phase advance is
# being used. UseInteger should be True. 
constr.add('beta', 15, refpts=len(half_arc), index=0) #beta_x at end of half_cell = 15m
constr.add('beta', 10, refpts=len(half_arc), index=1) # beta_y at end of half_cell = 10m
constr.add('dispersion', 0.6, refpts=len(half_arc), index=0) # dispersion_x at end of half cell is 0.6m
constr.add('dispersion', 0., refpts=len(half_arc), index=1, weight=1/10000) # dispersion_x prime at end of half cell is 0
constr.add('alpha', 0, refpts=len(half_arc), index=0, weight=1/1000) #alpha_x is 0 at end of cell
constr.add('alpha', 0, refpts=len(half_arc), index=1, weight=1/1000) #alpha_y is 0 at end of cell


# In[37]:


all_constraints = [constr]


# In[38]:


all_quad_vars = ['QF1','QD2', 'QD3', 'QF4'] #Quadrupole names that we want to add
all_sext_vars = ['SF','SD'] #Sextupole names that we want to do
all_vars = []

# For each quadrupole family, create an element variable for the strength 
# and add it to our list of variables
for var in all_quad_vars:
    all_vars.append(at.ElementVariable(at.get_uint32_index(half_arc, var), 'K'))
    
# Similarly for sextupole
for var in all_sext_vars:
    all_vars.append(at.ElementVariable(at.get_uint32_index(half_arc, var), 'H'))


# In[39]:


new_lat = at.match(half_arc, all_vars, all_constraints)


# In[40]:


new_lat.plot_beta(twiss_in=ld0)


# In[41]:


#get_ipython().run_line_magic('pinfo', 'constr.add')


# ## Dynamic Aperture

# In[ ]:





# We have several build in functions for the computation of the DA and lifetime. 

# In[42]:


#get_ipython().run_line_magic('pinfo', 'at.get_acceptance')


# In[43]:


ring.enable_6d()
bound, surv, track = at.get_acceptance(ring, ('x','y'), (11,11), (120e-3,60e-3), nturns=4000)


# In[44]:


plt.plot(*bound,marker='.', color='r')
plt.plot(*surv,marker='.',linestyle='None', color='b')
plt.plot(*track, color='g')
plt.show()


# In[ ]:





# In[ ]:




